package java.awt;

import java.security.AccessController;
import java.security.PrivilegedAction;
import sun.awt.PeerEvent;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.TimerTask;
import java.util.Timer;
import sun.util.logging.PlatformLogger;

class WaitDispatchSupport implements SecondaryLoop
{
    private static final PlatformLogger log;
    private EventDispatchThread dispatchThread;
    private EventFilter filter;
    private volatile Conditional extCondition;
    private volatile Conditional condition;
    private long interval;
    private static Timer timer;
    private TimerTask timerTask;
    private AtomicBoolean keepBlockingEDT;
    private AtomicBoolean keepBlockingCT;
    private AtomicBoolean afterExit;
    private final Runnable wakingRunnable;
    
    private static synchronized void initializeTimer() {
        if (WaitDispatchSupport.timer == null) {
            WaitDispatchSupport.timer = new Timer("AWT-WaitDispatchSupport-Timer", true);
        }
    }
    
    public WaitDispatchSupport(EventDispatchThread eventDispatchThread) {
        this(eventDispatchThread, null);
    }
    
    public WaitDispatchSupport(EventDispatchThread dispatchThread, Conditional extCondition) {
        this.keepBlockingEDT = new AtomicBoolean(false);
        this.keepBlockingCT = new AtomicBoolean(false);
        this.afterExit = new AtomicBoolean(false);
        this.wakingRunnable = new Runnable() {
            @Override
            public void run() {
                WaitDispatchSupport.log.fine("Wake up EDT");
                synchronized (WaitDispatchSupport.access$900()) {
                    WaitDispatchSupport.this.keepBlockingCT.set(false);
                    WaitDispatchSupport.access$900().notifyAll();
                }
                WaitDispatchSupport.log.fine("Wake up EDT done");
            }
        };
        if (dispatchThread == null) {
            throw new IllegalArgumentException("The dispatchThread can not be null");
        }
        this.dispatchThread = dispatchThread;
        this.extCondition = extCondition;
        this.condition = new Conditional() {
            @Override
            public boolean evaluate() {
                if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINEST)) {
                    WaitDispatchSupport.log.finest("evaluate(): blockingEDT=" + WaitDispatchSupport.this.keepBlockingEDT.get() + ", blockingCT=" + WaitDispatchSupport.this.keepBlockingCT.get());
                }
                int n = WaitDispatchSupport.this.extCondition == null || WaitDispatchSupport.this.extCondition.evaluate();
                if (!WaitDispatchSupport.this.keepBlockingEDT.get() || n == 0) {
                    if (WaitDispatchSupport.this.timerTask != null) {
                        WaitDispatchSupport.this.timerTask.cancel();
                        WaitDispatchSupport.this.timerTask = null;
                    }
                    return false;
                }
                else {
                    return true;
                }
            }
        };
        return;
    }
    
    public WaitDispatchSupport(EventDispatchThread eventDispatchThread, Conditional conditional, EventFilter filter, long interval) {
        this(eventDispatchThread, conditional);
        this.filter = filter;
        if (interval < 0L) {
            throw new IllegalArgumentException("The interval value must be >= 0");
        }
        this.interval = interval;
        if (interval != 0L) {
            initializeTimer();
        }
        return;
    }
    
    @Override
    public boolean enter() {
        if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINE)) {
            WaitDispatchSupport.log.fine("enter(): blockingEDT=" + this.keepBlockingEDT.get() + ", blockingCT=" + this.keepBlockingCT.get());
        }
        if (!this.keepBlockingEDT.compareAndSet(false, true)) {
            WaitDispatchSupport.log.fine("The secondary loop is already running, aborting");
            return false;
        }
        try {
            if (this.afterExit.get()) {
                WaitDispatchSupport.log.fine("Exit was called already, aborting");
                return false;
            }
            final Runnable runnable = new Runnable() {
                @Override
                public void run() {
                    WaitDispatchSupport.log.fine("Starting a new event pump");
                    if (WaitDispatchSupport.this.filter == null) {
                        WaitDispatchSupport.this.dispatchThread.pumpEvents(WaitDispatchSupport.this.condition);
                    }
                    else {
                        WaitDispatchSupport.this.dispatchThread.pumpEventsForFilter(WaitDispatchSupport.this.condition, WaitDispatchSupport.this.filter);
                    }
                }
            };
            Thread currentThread = Thread.currentThread();
            if (currentThread == this.dispatchThread) {
                if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINEST)) {
                    WaitDispatchSupport.log.finest("On dispatch thread: " + this.dispatchThread);
                }
                if (this.interval != 0L) {
                    if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINEST)) {
                        WaitDispatchSupport.log.finest("scheduling the timer for " + this.interval + " ms");
                    }
                    WaitDispatchSupport.timer.schedule(this.timerTask = new TimerTask() {
                        @Override
                        public void run() {
                            if (WaitDispatchSupport.this.keepBlockingEDT.compareAndSet(true, false)) {
                                WaitDispatchSupport.access$800(WaitDispatchSupport.this);
                            }
                        }
                    }, this.interval);
                }
                SequencedEvent currentSequencedEvent = KeyboardFocusManager.getCurrentKeyboardFocusManager().getCurrentSequencedEvent();
                if (currentSequencedEvent != null) {
                    if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINE)) {
                        WaitDispatchSupport.log.fine("Dispose current SequencedEvent: " + currentSequencedEvent);
                    }
                    currentSequencedEvent.dispose();
                }
                AccessController.doPrivileged(new PrivilegedAction<Void>() {
                    @Override
                    public Void run() {
                        runnable.run();
                        return null;
                    }
                });
            }
            else {
                if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINEST)) {
                    WaitDispatchSupport.log.finest("On non-dispatch thread: " + currentThread);
                }
                this.keepBlockingCT.set(true);
                synchronized (getTreeLock()) {
                    if (this.afterExit.get()) {
                        boolean n = false;
                        return n;
                    }
                    if (this.filter != null) {
                        this.dispatchThread.addEventFilter(this.filter);
                    }
                    try {
                        this.dispatchThread.getEventQueue().postEvent(new PeerEvent(this, runnable, 1L));
                        if (this.interval > 0L) {
                            long currentTimeMillis = System.currentTimeMillis();
                            while (this.keepBlockingCT.get() && (this.extCondition == null || this.extCondition.evaluate()) && currentTimeMillis + this.interval > System.currentTimeMillis()) {
                                getTreeLock().wait(this.interval);
                            }
                        }
                        else {
                            while (this.keepBlockingCT.get() && (this.extCondition == null || this.extCondition.evaluate())) {
                                getTreeLock().wait();
                            }
                        }
                        if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINE)) {
                            WaitDispatchSupport.log.fine("waitDone " + this.keepBlockingEDT.get() + " " + this.keepBlockingCT.get());
                        }
                    }
                    catch (InterruptedException obj) {
                        if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINE)) {
                            WaitDispatchSupport.log.fine("Exception caught while waiting: " + obj);
                        }
                    }
                    finally {
                        if (this.filter != null) {
                            this.dispatchThread.removeEventFilter(this.filter);
                        }
                    }
                }
            }
            return true;
        }
        finally {
            this.keepBlockingEDT.set(false);
            this.keepBlockingCT.set(false);
            this.afterExit.set(false);
        }
    }
    
    @Override
    public boolean exit() {
        if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINE)) {
            WaitDispatchSupport.log.fine("exit(): blockingEDT=" + this.keepBlockingEDT.get() + ", blockingCT=" + this.keepBlockingCT.get());
        }
        this.afterExit.set(true);
        if (this.keepBlockingEDT.getAndSet(false)) {
            this.wakeupEDT();
            return true;
        }
        return false;
    }
    
    private static final Object getTreeLock() {
        return Component.LOCK;
    }
    
    private void wakeupEDT() {
        if (WaitDispatchSupport.log.isLoggable(PlatformLogger.Level.FINEST)) {
            WaitDispatchSupport.log.finest("wakeupEDT(): EDT == " + this.dispatchThread);
        }
        this.dispatchThread.getEventQueue().postEvent(new PeerEvent(this, this.wakingRunnable, 1L));
    }
    
    static void access$800(WaitDispatchSupport waitDispatchSupport) {
        waitDispatchSupport.wakeupEDT();
    }
    
    static Object access$900() {
        return getTreeLock();
    }
    
    static {
        log = PlatformLogger.getLogger("java.awt.event.WaitDispatchSupport");
    }
}
